package edu.uiuc.dsi.mutex;

import java.util.concurrent.Semaphore;

/**
 * An abstract class providing common functionality to all mutual exclusion objects.
 * 
 * @author Wade Fagen
 * @since 0.0.1.2
 */
public abstract class Base extends edu.uiuc.dsi.Base
{
	public static final byte NOT_REQUESTING_OR_IN_CRITICAL_SECTION = 0x01;
	public static final byte REQUESTING_CRITICAL_SECTION = 0x02;
	public static final byte EXECUTING_CRITICAL_SECTION = 0x03;

	private byte currentMutualExclusionState = NOT_REQUESTING_OR_IN_CRITICAL_SECTION;
	private Semaphore objSemaphoreLock = new Semaphore(0);

	/**
	 * A constructor to create a <code>edu.uiuc.dsi.mutex</code> object with a given name. 
	 * 
	 * @param componentName The user-defined name of the <code>edu.uiuc.dsi.mutex</code> object.
	 */
	public Base(String componentName)
	{
		super(componentName);
	}

	protected abstract void doRequestMutExTasks();
	protected abstract void doReleaseMutExTasks();
	protected abstract void doBeginMutExTasks();
	

	/**
	 * Initializes child controls of the object. Currently, objects derived from edu.uiuc.dsi.mutex.Base do not
	 * support child controls.
	 */
	@Override
	public void initializeChildControls()
	{
		// mutex functions don't have child controls.
	}
	
	
	/**
	 * Requests a critical section. This call will block the current thread until a critical section
	 * is acquired, the thread is interrupted, or the mutual exclusion object enters into an unsafe state.
	 * 
	 * @throws MutExException A <code>MutExException</code> is thrown if requesting the critical section will place
	 * the mutual exclusion object into an unsafe state.
	 *  
	 * @throws InterruptedException 
	 */
	public void requestMutEx() throws MutExException, InterruptedException
	{
		switch (currentMutualExclusionState)
		{
			case NOT_REQUESTING_OR_IN_CRITICAL_SECTION:
				currentMutualExclusionState = REQUESTING_CRITICAL_SECTION;
				
				doRequestMutExTasks();
				objSemaphoreLock.acquire();
				
				break;
				
			case REQUESTING_CRITICAL_SECTION:
				throw new MutExException("A second request to the same critical section cannot be made when the first request has not been granted.");
				
			case EXECUTING_CRITICAL_SECTION:
				throw new MutExException("A second request to the same critical section cannot be made when the first request is still executing the critical section.");
		}
	}
	
	/**
	 * Notifies the mutual exclusion object that an a calling thread is safe to enter into its requested critical section.
	 * 
	 * @throws MutExException A <code>MutExException</code> is thrown if requesting the critical section will place
	 * the mutual exclusion object into an unsafe state.
	 */
	protected void beginMutEx() throws MutExException
	{
		switch (currentMutualExclusionState)
		{
			case NOT_REQUESTING_OR_IN_CRITICAL_SECTION:
				throw new MutExException("A critical section cannot begin without having been requested.");
				
			case REQUESTING_CRITICAL_SECTION:
				currentMutualExclusionState = EXECUTING_CRITICAL_SECTION;
				doBeginMutExTasks();
				objSemaphoreLock.release();

				break;
				
			case EXECUTING_CRITICAL_SECTION:
				throw new MutExException("A critical section cannot begin when a critical section is already being executed.");
		}
	}
	
	/**
	 * Releases the resources used by the critical section. This method must be called after a successful call of {@link #requestMutEx()}
	 * to release the critical section or all other objects requesting the critical section will deadlock waiting on the release of the
	 * critical section.
	 * 
	 * @throws MutExException A <code>MutExException</code> is thrown if requesting the critical section will place
	 * the mutual exclusion object into an unsafe state.
	 */
	public void releaseMutEx() throws MutExException
	{
		switch (currentMutualExclusionState)
		{
			case NOT_REQUESTING_OR_IN_CRITICAL_SECTION:
			case REQUESTING_CRITICAL_SECTION:
				throw new MutExException("A mutual exclusion object cannot be released when it is not in its critical section.");
				
			case EXECUTING_CRITICAL_SECTION:
				doReleaseMutExTasks();
				currentMutualExclusionState = NOT_REQUESTING_OR_IN_CRITICAL_SECTION;
				break;
		}
	}
	
	/**
	 * Get the current enumerated state of the mutual exclusion object.
	 * 
	 * @return The enumerated state of the mutual exclusion object.
	 */
	public byte getCurrentMutualExclusionState()
	{
		return currentMutualExclusionState;
	}
}
